import importlib.util as imputil
import inspect
from pathlib import Path
from typing_extensions import Annotated

import typer

from brij.utils.server.base import BaseServer

app = typer.Typer(help="Brij CLI for servers with ZMQ")


def load_module(file_path: Path):
    spec = imputil.spec_from_file_location(file_path.stem, str(file_path))
    mod = imputil.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(mod)  # type: ignore
    return mod


def find_server_class(module) -> type[BaseServer]:
    for _, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, BaseServer) and obj is not BaseServer:
            return obj
    raise ValueError("❌ No subclass of BaseServer found in the module.")


@app.command("hello")
def hello():
    print("Welcome to Brij")


@app.command(name="serve")
def serve(
    server_file: Annotated[
        Path, typer.Argument(help="Path to the server implementation file")
    ],
    port: Annotated[
        int, typer.Option("--port", "-p", help="Port to bind the server to")
    ] = 5555,
):
    """
    Serve a ZMQ server from a given Python file.
    """
    typer.echo(f"🔍 Loading server from {server_file}...")
    module = load_module(server_file)
    server_class = find_server_class(module)

    typer.echo(f"🚀 Starting {server_class.__name__} on port {port}...")
    server_instance = server_class(port=port)
    try:
        server_instance.run()
    except KeyboardInterrupt:
        print("Shutting down server...")
        server_instance.close()



from typing import Generic, TypeVar

import zmq

from brij.utils.msg import deserialize, serialize

TInput = TypeVar("TInput", bound=dict)
TOutput = TypeVar("TOutput", bound=dict)


class Client(Generic[TInput, TOutput]):
    """
    Base class for a client.
    """

    def __init__(self, host: str = "localhost", port: int = 5555) -> None:
        super().__init__()

        self._context: zmq.Context[zmq.Socket[bytes]] = zmq.Context()
        self._socket: zmq.Socket[bytes] = self._context.socket(socket_type=zmq.REQ)
        self._host, self._port = host, port

    def connect(self) -> bool:
        try:
            self._socket.connect(addr=f"tcp://{self._host}:{self._port}")
        except Exception:
            return False
        else:
            return True

    def __call__(self, msg: TInput) -> TOutput:
        self._socket.send(serialize(msg=msg))  # type: ignore
        return deserialize(msg_bytes=self._socket.recv())  # type: ignore



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

import zmq

from brij.utils.msg import deserialize, serialize

TInput = TypeVar("TInput")
TOutput = TypeVar("TOutput")


class BaseServer(Generic[TInput, TOutput], ABC):
    """
    Base server.
    """

    def __init__(self, port: int = 5555) -> None:
        super().__init__()
        self._context = zmq.Context[zmq.Socket[bytes]]()
        self._socket: zmq.Socket[bytes] = self._context.socket(socket_type=zmq.REP)
        self._socket.bind(f"tcp://*:{port}")

    def _send(self, msg: TOutput) -> None:
        self._socket.send(serialize(msg=msg))  # type: ignore

    def _handle_request(self, msg: bytes) -> None:
        msg_orig: TOutput = deserialize(msg_bytes=msg)  # type: ignore
        response_msg = self.process(msg=msg_orig)  # type: ignore
        self._send(msg=response_msg)

    def run(self) -> None:
        while True:
            msg = self._socket.recv()
            self._handle_request(msg=msg)

    @abstractmethod
    def process(self, msg: TInput) -> TOutput:
        pass

    def close(self) -> None:
        self._socket.close()
        self._context.term()



import blosc
import msgpack
import msgpack_numpy as mnp

mnp.patch()


def serialize(msg=dict) -> bytes:
    msg_packed = msgpack.packb(msg, default=mnp.encode)
    msg_compressed = blosc.compress(msg_packed)
    return msg_compressed


def deserialize(msg_bytes: bytes) -> dict:
    decompressed = blosc.decompress(msg_bytes)
    msg_decoded = msgpack.unpackb(decompressed, object_hook=mnp.decode)
    return msg_decoded
